#include <avr/io.h>

/* Pins */
#define LED_4 13
#define LED_3 12
#define LED_2 11
#define LED_1 10
#define RGB_R 9
#define RGB_G 8
#define RGB_B 7
#define BUZZER 6
#define BTN_4 5
#define BTN_3 4
#define BTN_2 3
#define BTN_1 2

/* Button debounce check */
#define DEBOUNCE_CT 300
unsigned long t_debounce_1 = 0;

bool valid_press() {
  unsigned long t_int = millis();
  if (t_int - t_debounce_1 > DEBOUNCE_CT) {
    t_debounce_1 = t_int;
    return true;
  }
  return false;
}

/* Timed moves */
#define TIME_PER_MOVE 3000
unsigned long t_move = 0;

/* States */
#define STANDBY 0
#define SELECT 1
#define START 2
#define EASY 3
#define MEDIUM 4
#define HARD 5
#define HARD_T 6
#define PLAYING 7
#define LOSE 10
#define WIN 11
volatile byte state = STANDBY;

void setup() {
  pinMode(RGB_R, OUTPUT);
  pinMode(RGB_G, OUTPUT);
  pinMode(RGB_B, OUTPUT);
  pinMode(LED_1, OUTPUT);
  pinMode(LED_2, OUTPUT);
  pinMode(LED_3, OUTPUT);
  pinMode(LED_4, OUTPUT);
  pinMode(BUZZER, OUTPUT);
  pinMode(BTN_1, INPUT_PULLUP);
  pinMode(BTN_2, INPUT_PULLUP);
  pinMode(BTN_3, INPUT_PULLUP);
  pinMode(BTN_4, INPUT_PULLUP);
  
  cli();
  PCICR |= (1 << PCIE2);
  PCIFR |= (1 << PCIF2);
  PCMSK2 |= (1 << PCINT18);
  PCMSK2 |= (1 << PCINT19);
  PCMSK2 |= (1 << PCINT20);
  PCMSK2 |= (1 << PCINT21);
  sei();
}

volatile int last_pressed = 0;

#define MAX_LENGTH 10
volatile byte cl_arr[MAX_LENGTH];
volatile byte cl_index;
volatile byte cl_len;

void level(int length, bool timed = false);

void loop() {
  if (state == STANDBY) {
    digitalWrite(LED_1, HIGH);
    digitalWrite(LED_2, HIGH);
    digitalWrite(LED_3, HIGH);
    digitalWrite(LED_4, HIGH);
  } else {
    digitalWrite(LED_1, LOW);
    digitalWrite(LED_2, LOW);
    digitalWrite(LED_3, LOW);
    digitalWrite(LED_4, LOW);
    delay(300);
    if (state == EASY) {
      level(2);
    } else if (state == MEDIUM) {
      level(3);
    } else if (state == HARD) {
      level(4);
    } else if (state == HARD_T) {
      level(4, true);
    }
  }
}

void turn_off_all() {
  digitalWrite(LED_1, LOW);
  digitalWrite(LED_2, LOW);
  digitalWrite(LED_3, LOW);
  digitalWrite(LED_4, LOW);
}

void level(int length, bool timed)
{
  digitalWrite(RGB_B, HIGH);
  delay(300);
  state = START;
  cl_index = 0;
  cl_len = length;
  /* Generate sequence */
  for (unsigned int i = 0; i < length; i++) {
    cl_arr[i] = random(LED_1, LED_4 + 1);
    digitalWrite(cl_arr[i], HIGH);
    delay(1000);
    digitalWrite(cl_arr[i], LOW);
    delay(700);
  }
  state = PLAYING;
  digitalWrite(RGB_B, LOW);
  digitalWrite(RGB_G, HIGH);
  t_move = millis();
  while (state == PLAYING) {
    if (timed) {
      unsigned int t_c = millis();
      if (t_c - t_move > TIME_PER_MOVE) {
        state = LOSE;
      }
    }
    if (last_pressed > 0) {  // Feedback
      digitalWrite(last_pressed, HIGH);
    }
  }
  digitalWrite(RGB_G, LOW);
  turn_off_all();
  if (state == LOSE) {
    fail();
  } else {  // Feedback for the last pressed button
    digitalWrite(last_pressed, HIGH);
    delay(300);
    digitalWrite(last_pressed, LOW);
    win();
  }
}

ISR(PCINT2_vect) {
  if (state == PLAYING) {  // Accept player input
    if (((PIND & (1 << PD2)) == 0) && valid_press()) {  // BTN_1
      if (cl_arr[cl_index] == LED_1) {
        cl_index++;
        t_move = millis();
      } else {
        state = LOSE;
      }
    } else if (((PIND & (1 << PD3)) == 0) && valid_press()) {  // BTN_2
      if (cl_arr[cl_index] == LED_2) {
        cl_index++;
        t_move = millis();
      } else {
        state = LOSE;
      }
    } else if (((PIND & (1 << PD4)) == 0) && valid_press()) {  // BTN_3  
      if (cl_arr[cl_index] == LED_3) {
        cl_index++;
        t_move = millis();
      } else {
        state = LOSE;
      }
    } else if (((PIND & (1 << PD5)) == 0) && valid_press()) {  // BTN_4
      if (cl_arr[cl_index] == LED_4) {
        cl_index++;
        t_move = millis();
      } else {
        state = LOSE;
      }
    }
    if (cl_index == cl_len) {  // Count corect choices
      state = WIN;
    }
    if (state != LOSE) {  // Feedback
      if (cl_index >= 2) {
        digitalWrite(cl_arr[cl_index - 2], LOW);
      }
      last_pressed = cl_arr[cl_index - 1];
    }
  } else if (state == STANDBY && ((PIND & (1 << PD5)) == 0) && valid_press()) {  // Enter Select menu
    state = SELECT;
  } else if (state == SELECT && ((PIND & (1 << PD2)) == 0) && valid_press()) {  // Easy
    state = EASY;
  } else if (state == SELECT && ((PIND & (1 << PD3)) == 0) && valid_press()) {  // Medium
    state = MEDIUM;
  } else if (state == SELECT && ((PIND & (1 << PD4)) == 0) && valid_press()) {  // Hard
    state = HARD;
  } else if (state == SELECT && ((PIND & (1 << PD5)) == 0) && valid_press()) {  // Hard with time
    state = HARD_T;
  }
}

void win() {
  digitalWrite(RGB_G, HIGH);
  delay(300);
  digitalWrite(RGB_G, LOW);
  delay(300);
  digitalWrite(RGB_G, HIGH);
  delay(300);
  digitalWrite(RGB_G, LOW);
  delay(300);
  digitalWrite(RGB_G, HIGH);
  delay(300);
  digitalWrite(RGB_G, LOW);
  delay(300);
  sing_win();
  state = STANDBY;
}

void fail() {
  digitalWrite(RGB_R, HIGH);
  delay(300);
  digitalWrite(RGB_R, LOW);
  delay(300);
  digitalWrite(RGB_R, HIGH);
  delay(300);
  digitalWrite(RGB_R, LOW);
  delay(300);
  digitalWrite(RGB_R, HIGH);
  delay(300);
  digitalWrite(RGB_R, LOW);
  delay(300);
  sing_lose();
  state = STANDBY;
}

/* BUZZER songs */
int buzzer = BUZZER;
#define NOTE_B0  31
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62
#define NOTE_C2  65
#define NOTE_CS2 69
#define NOTE_D2  73
#define NOTE_DS2 78
#define NOTE_E2  82
#define NOTE_F2  87
#define NOTE_FS2 93
#define NOTE_G2  98
#define NOTE_GS2 104
#define NOTE_A2  110
#define NOTE_AS2 117
#define NOTE_B2  123
#define NOTE_C3  131
#define NOTE_CS3 139
#define NOTE_D3  147
#define NOTE_DS3 156
#define NOTE_E3  165
#define NOTE_F3  175
#define NOTE_FS3 185
#define NOTE_G3  196
#define NOTE_GS3 208
#define NOTE_A3  220
#define NOTE_AS3 233
#define NOTE_B3  247
#define NOTE_C4  262
#define NOTE_CS4 277
#define NOTE_D4  294
#define NOTE_DS4 311
#define NOTE_E4  330
#define NOTE_F4  349
#define NOTE_FS4 370
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define NOTE_C6  1047
#define NOTE_CS6 1109
#define NOTE_D6  1175
#define NOTE_DS6 1245
#define NOTE_E6  1319
#define NOTE_F6  1397
#define NOTE_FS6 1480
#define NOTE_G6  1568
#define NOTE_GS6 1661
#define NOTE_A6  1760
#define NOTE_AS6 1865
#define NOTE_B6  1976
#define NOTE_C7  2093
#define NOTE_CS7 2217
#define NOTE_D7  2349
#define NOTE_DS7 2489
#define NOTE_E7  2637
#define NOTE_F7  2794
#define NOTE_FS7 2960
#define NOTE_G7  3136
#define NOTE_GS7 3322
#define NOTE_A7  3520
#define NOTE_AS7 3729
#define NOTE_B7  3951
#define NOTE_C8  4186
#define NOTE_CS8 4435
#define NOTE_D8  4699
#define NOTE_DS8 4978
#define REST      0

int tempo_win = 158;
int tempo_lose = 160;

int melody_win[] = {
  NOTE_AS4,8, NOTE_AS4,8, NOTE_AS4,8,//1
  NOTE_F5,2, NOTE_C6,2,
  NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F6,2, NOTE_C6,4,  
  NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F6,2, NOTE_C6,4,  
  NOTE_AS5,8, NOTE_A5,8, NOTE_AS5,8, NOTE_G5,2, NOTE_C5,8, NOTE_C5,8, NOTE_C5,8,
  NOTE_F5,2, NOTE_C6,2,
  NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F6,2, NOTE_C6,4,  
  
  NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F6,2, NOTE_C6,4, //8  
  NOTE_AS5,8, NOTE_A5,8, NOTE_AS5,8, NOTE_G5,2, NOTE_C5,-8, NOTE_C5,16, 
  NOTE_D5,-4, NOTE_D5,8, NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F5,8,
  NOTE_F5,8, NOTE_G5,8, NOTE_A5,8, NOTE_G5,4, NOTE_D5,8, NOTE_E5,4,NOTE_C5,-8, NOTE_C5,16,
  NOTE_D5,-4, NOTE_D5,8, NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F5,8,
  
  NOTE_C6,-8, NOTE_G5,16, NOTE_G5,2, REST,8, NOTE_C5,8,//13
  NOTE_D5,-4, NOTE_D5,8, NOTE_AS5,8, NOTE_A5,8, NOTE_G5,8, NOTE_F5,8,
  NOTE_F5,8, NOTE_G5,8, NOTE_A5,8, NOTE_G5,4, NOTE_D5,8, NOTE_E5,4,NOTE_C6,-8, NOTE_C6,16,
  NOTE_F6,4, NOTE_DS6,8, NOTE_CS6,4, NOTE_C6,8, NOTE_AS5,4, NOTE_GS5,8, NOTE_G5,4, NOTE_F5,8,
  NOTE_C6,1
};
int melody_lose[] = {
  NOTE_A4,-4, NOTE_A4,-4, NOTE_A4,16, NOTE_A4,16, NOTE_A4,16, NOTE_A4,16, NOTE_F4,8, REST,8,
  NOTE_A4,-4, NOTE_A4,-4, NOTE_A4,16, NOTE_A4,16, NOTE_A4,16, NOTE_A4,16, NOTE_F4,8, REST,8,
  NOTE_A4,4, NOTE_A4,4, NOTE_A4,4, NOTE_F4,-8, NOTE_C5,16,

  NOTE_A4,4, NOTE_F4,-8, NOTE_C5,16, NOTE_A4,2,//4
  NOTE_E5,4, NOTE_E5,4, NOTE_E5,4, NOTE_F5,-8, NOTE_C5,16,
  NOTE_A4,4, NOTE_F4,-8, NOTE_C5,16, NOTE_A4,2,
  
  NOTE_A5,4, NOTE_A4,-8, NOTE_A4,16, NOTE_A5,4, NOTE_GS5,-8, NOTE_G5,16, //7 
  NOTE_DS5,16, NOTE_D5,16, NOTE_DS5,8, REST,8, NOTE_A4,8, NOTE_DS5,4, NOTE_D5,-8, NOTE_CS5,16,

  NOTE_C5,16, NOTE_B4,16, NOTE_C5,16, REST,8, NOTE_F4,8, NOTE_GS4,4, NOTE_F4,-8, NOTE_A4,-16,//9
  NOTE_C5,4, NOTE_A4,-8, NOTE_C5,16, NOTE_E5,2,

  NOTE_A5,4, NOTE_A4,-8, NOTE_A4,16, NOTE_A5,4, NOTE_GS5,-8, NOTE_G5,16, //7 
  NOTE_DS5,16, NOTE_D5,16, NOTE_DS5,8, REST,8, NOTE_A4,8, NOTE_DS5,4, NOTE_D5,-8, NOTE_CS5,16,

  NOTE_C5,16, NOTE_B4,16, NOTE_C5,16, REST,8, NOTE_F4,8, NOTE_GS4,4, NOTE_F4,-8, NOTE_A4,-16,//9
  NOTE_A4,4, NOTE_F4,-8, NOTE_C5,16, NOTE_A4,2
};
int notes_win = sizeof(melody_win) / sizeof(melody_win[0]) / 2;
int wholenote_win = (60000 * 4) / tempo_win;
int divider_win = 0, noteDuration_win = 0;
int notes_lose = sizeof(melody_lose) / sizeof(melody_lose[0]) / 2;
int wholenote_lose = (60000 * 4) / tempo_lose;
int divider_lose = 0, noteDuration_lose = 0;

void sing_win() {
  for (int thisNote = 0; thisNote < notes_win * 2; thisNote = thisNote + 2) {
    divider_win = melody_win[thisNote + 1];
    if (divider_win > 0) {
      noteDuration_win = (wholenote_win) / divider_win;
    } else if (divider_win < 0) {
      noteDuration_win = (wholenote_win) / abs(divider_win);
      noteDuration_win *= 1.5;
    }
    tone(buzzer, melody_win[thisNote], noteDuration_win*0.9);
    delay(noteDuration_win);
    noTone(buzzer);
  }
}
void sing_lose() {
  for (int thisNote = 0; thisNote < notes_lose * 2; thisNote = thisNote + 2) {
    divider_lose = melody_lose[thisNote + 1];
    if (divider_lose > 0) {
      noteDuration_lose = (wholenote_lose) / divider_lose;
    } else if (divider_lose < 0) {
      noteDuration_lose = (wholenote_lose) / abs(divider_lose);
      noteDuration_lose *= 1.5;
    }
    tone(buzzer, melody_lose[thisNote], noteDuration_lose*0.9);
    delay(noteDuration_lose);
    noTone(buzzer);
  }
}
